//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Written by Philippe Leefsma - Autodesk Developer Technical Services. 2010.
// 
// Feedback and bug report: Philippe.Leefsma@Autodesk.com
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#include "StdAfx.h"
#include "SurfaceListDlg.h"
#include "AcDbAssocVariable.h"
#include "AcDbAssocActionBody.h"
#include "AcDbAssocDependency.h"
#include "AcDbAssocPersSubentIdPE.h"
#include "AcDbAssocValueDependency.h"
#include "AcDbAssoc2dConstraintGroup.h" 
#include "AcDbAssocPathBasedSurfaceActionBody.h"
#include "AcDbAssocParamBasedActionBody.h"
#include "AcDbAssocPathActionParam.h"
#include "ConstrainDbgHelper.h"

IMPLEMENT_DYNAMIC(CSurfaceDbgDlg, CAdUiDialog)

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CSurfaceDbgDlg::CSurfaceDbgDlg(CWnd* pParent /*=NULL*/)
	: CAdUiDialog(CSurfaceDbgDlg::IDD, pParent)
{
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CSurfaceDbgDlg::~CSurfaceDbgDlg()
{
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CSurfaceDbgDlg::OnInitDialog()
{
	CAdUiDialog::OnInitDialog();

	AcDbDatabase *pDatabase = acdbHostApplicationServices()->workingDatabase();
	
	AcDbObjectPointer<AcDbBlockTable> blockTable(pDatabase->blockTableId(), kForRead);

	AcDbBlockTableIterator* pIterator = NULL;
	blockTable->newIterator(pIterator);

	AcDbObjectId Id;
	const TCHAR* pName;

	while(!pIterator->done())
	{
		pIterator->getRecordId(Id);

		AcDbObjectPointer<AcDbBlockTableRecord> btr(Id, kForRead);
		btr->getName(pName);

		m_ComboBoxCtrl.AddString(pName);
		pIterator->step();
	}

	delete pIterator;

	m_ComboBoxCtrl.SetCurSel(0);

	m_TreeCtrl.SetItemHeight(18);

	InitializeListCtrl();

	OnComboSelectionChanged();

	return TRUE;  
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSurfaceDbgDlg::InitializeListCtrl()
{
	CRect lvRect;
    int col1_width, col2_width;
    m_ListCtrl.GetClientRect(lvRect);

    col1_width = lvRect.Width() / 2;
    col2_width = lvRect.Width() - col1_width - ::GetSystemMetrics(SM_CXVSCROLL);

    LV_COLUMN lvc;
    lvc.mask = LVCF_WIDTH | LVCF_SUBITEM | LVCF_TEXT;
    lvc.cx = col1_width;
    lvc.pszText = _T("Field");
    lvc.iSubItem = 0;
    m_ListCtrl.InsertColumn(0, &lvc);

    lvc.cx = col2_width;
    lvc.iSubItem = 1;
    lvc.pszText = _T("Value");
    m_ListCtrl.InsertColumn(1, &lvc);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Use: We need to cleanup the "ItemReference*" attached to the TreeItems
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSurfaceDbgDlg::Cleanup()
{
	map <HTREEITEM, AcDbObjectId>::iterator iterator;

	int s = m_ItemToObjectIdMap.size();

	for(iterator = m_ItemToObjectIdMap.begin(); iterator != m_ItemToObjectIdMap.end() ; ++iterator)
	{
		DWORD_PTR data = m_TreeCtrl.GetItemData(iterator->first);

		delete (ItemReference*)data;
	}

	m_ItemToObjectIdMap.clear();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSurfaceDbgDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_COMBO, m_ComboBoxCtrl);
	DDX_Control(pDX, IDC_TREE, m_TreeCtrl);
	DDX_Control(pDX, IDC_LIST, m_ListCtrl);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BEGIN_MESSAGE_MAP(CSurfaceDbgDlg, CDialog)
	ON_NOTIFY(TVN_SELCHANGED, IDC_TREE, &CSurfaceDbgDlg::OnTreeviewSelectionChanged)
	ON_CBN_SELCHANGE(IDC_COMBO, &CSurfaceDbgDlg::OnComboSelectionChanged)
	ON_BN_CLICKED(IDOK, &CSurfaceDbgDlg::OnBnClickedOk)
	ON_WM_CLOSE()
END_MESSAGE_MAP()

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSurfaceDbgDlg::OnComboSelectionChanged()
{
	Cleanup();

	m_TreeCtrl.DeleteAllItems();
	m_ListCtrl.DeleteAllItems();

	ACHAR buff[1024];

	int nIndex = m_ComboBoxCtrl.GetCurSel();
	m_ComboBoxCtrl.GetLBText(nIndex, buff);

	AcDbDatabase *pDatabase = acdbHostApplicationServices()->workingDatabase();
	
	AcDbObjectPointer<AcDbBlockTable> blockTable(pDatabase->blockTableId(), kForRead);

	AcDbObjectId spaceId;
	blockTable->getAt(buff, spaceId);

	vector<AcDbObjectId> surfaceIds;
	m_surfaceUtility.getAllSurfaces(spaceId, surfaceIds);

	vector<AcDbObjectId>::iterator iterator;
	for(iterator = surfaceIds.begin(); iterator != surfaceIds.end(); ++iterator)
	{
		addSurfaceTreeItem(*iterator);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSurfaceDbgDlg::OnTreeviewSelectionChanged(NMHDR *pNMHDR, LRESULT *pResult)
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
      
	HTREEITEM hItem = pNMTreeView->itemNew.hItem;

	if(m_ItemToObjectIdMap.find(hItem) == m_ItemToObjectIdMap.end())
		return;

	vector<ObjectProperty> properties;

	AcDbObjectId id = m_ItemToObjectIdMap[hItem];

	if(id != AcDbObjectId::kNull)
	{
		m_surfaceUtility.getProperties(id, properties);
	}
	else
	{ 
		//Non Db resident objects are referenced using "ItemReference*"
		if(pNMTreeView->itemNew.lParam != 0)
		{
			m_surfaceUtility.getProperties((ItemReference*)pNMTreeView->itemNew.lParam, properties);
		}
	}

	displayPropertyList(properties);

	*pResult = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSurfaceDbgDlg::OnBnClickedOk()
{
	Cleanup();

	CAdUiDialog::OnOK();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSurfaceDbgDlg::OnClose()
{
	Cleanup();

	CAdUiDialog::OnClose();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Use: Add a new TreeItem in the Tree Control
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
HTREEITEM CSurfaceDbgDlg::addTreeItem(LPCTSTR name, HTREEITEM parent, const AcDbObjectId& id, ItemReference* pItemRef)
{
	TV_ITEM tvItem;
    
    tvItem.mask = TVIF_TEXT | TVIF_PARAM;
    tvItem.pszText = const_cast<TCHAR*>(name);
    tvItem.cchTextMax = lstrlen(name);
    tvItem.lParam = (LPARAM)0;

	TV_INSERTSTRUCT tvInsert;

    tvInsert.item = tvItem;
	tvInsert.hInsertAfter = TVI_LAST;
    tvInsert.hParent = parent;

	HTREEITEM hItem = m_TreeCtrl.InsertItem(&tvInsert);

	//Insert the new item in our Handle-ObjectId Map.
	m_ItemToObjectIdMap[hItem] = id;

	//However the ObjectId can be null (in case of non db-resident entity).
	//So we also attach a pointer to an ItemReference associated with this TreeItem.
	//In case of a db-resident entity, pItemRef will be null but it doesn't matter.
	m_TreeCtrl.SetItemData(hItem, (DWORD_PTR)pItemRef);

    return hItem;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Use: Add a Surface Entity to the TreeList and build the sub items hierachy
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
HTREEITEM CSurfaceDbgDlg::addSurfaceTreeItem(const AcDbObjectId& surfaceId)
{
	Acad::ErrorStatus es;

	AcDbObjectPointer <AcDbSurface> surface(surfaceId, AcDb::kForRead);

	if(surface.openStatus() != Acad::eOk)
	{
		return 0;
	}

	//Add the "root" surface entity item
	CString str;
	HTREEITEM hSurfaceItem = addTreeItem(CConstrainDbgHelper::objToClassAndHandleStr(surface.object(), str), TVI_ROOT, surfaceId);

	//Check if the surface is associative, in case it isn't "creationActionBodyId" is Null
	if(surface->creationActionBodyId().isNull())
	{
		return hSurfaceItem;
	}

	//Add ActionBody item
	AcDbObjectPointer <AcDbAssocActionBody> actionBody(surface->creationActionBodyId(), AcDb::kForRead);

	HTREEITEM hActionBodyItem = addTreeItem(CConstrainDbgHelper::objToClassAndHandleStr(actionBody.object(), str), hSurfaceItem, surface->creationActionBodyId());

	//Check if ActionBody is "path-based". Not all surfaces are path-based.
	if (actionBody->isKindOf(AcDbAssocPathBasedSurfaceActionBody::desc()))
	{
		AcDbAssocPathBasedSurfaceActionBody* pathBasedSurfaceActionBody = AcDbAssocPathBasedSurfaceActionBody::cast(actionBody);

		AcDbObjectIdArray pathParamIds;
		es = pathBasedSurfaceActionBody->getInputPathParams(pathParamIds);

		for(int i=0; i<pathParamIds.length(); ++i)
		{
			AcDbObjectPointer <AcDbAssocPathActionParam> param(pathParamIds[i], AcDb::kForRead);

			HTREEITEM hPathItem = addTreeItem(CConstrainDbgHelper::objToClassAndHandleStr(param.object(), str), hActionBodyItem, pathParamIds[i]);

			AcArray<AcArray<AcDbEdgeRef>> edgeRefs;
			es = param->getEdgeRefArray(edgeRefs);

			//Add "AcDbEdgeRef" items to the Tree. 
			//AcDbEdgeRef is not a db-resident entity, 
			//so we can't use an ObjectId to retrieve it's properties if selected in the tree.
			//That's why an "EdgeRefReference" is created to identify it later if needed.

			for(int j=0; j<edgeRefs.length(); ++j)
			{
				for(int k=0; k<edgeRefs[j].length(); ++k)
				{
					AcDbEdgeRef edgeRef = edgeRefs[j][k];

					EdgeRefReference* itemRef = new EdgeRefReference(pathParamIds[i], j, k);
					
					HTREEITEM hEdgeRefItem = addTreeItem(L"<AcDbEdgeRef>", hPathItem, AcDbObjectId::kNull, itemRef);
				}
			}
		}
	}

	//Check if ActionBody is "param-based". Just a precaution, since every ActionBody should be param-based.
	if (actionBody->isKindOf(AcDbAssocParamBasedActionBody::desc()))
	{
		AcDbAssocParamBasedActionBody* paramBasedActionBody = AcDbAssocParamBasedActionBody::cast(actionBody);

		AcDbObjectIdArray ownedParamIds = paramBasedActionBody->ownedParams();

		for(int i = 0; i<ownedParamIds.length(); ++i)
		{
			AcDbObjectPointer <AcDbAssocActionParam> param(ownedParamIds[i], AcDb::kForRead);
		
			//Check it is not an AcDbAssocPathActionParam (already in the TreeCtrl otherwise, so skip it).
			if(AcDbAssocPathActionParam::cast(param.object()))
				continue;

			HTREEITEM hParamItem = addTreeItem(CConstrainDbgHelper::objToClassAndHandleStr(param.object(), str), hActionBodyItem, ownedParamIds[i]);
		}
	}
	
	//Add ActionBody dependencies in the Tree. Every ActionBody should have - at least - a dependency (since it is associative hehe:))
	AcDbObjectIdArray dependencyIds;
	if((es = actionBody->getDependencies(true, true, dependencyIds)) == Acad::eOk)
	{
		for(int i = 0; i<dependencyIds.length(); ++i)
		{
			AcDbObjectPointer <AcDbAssocDependency> assocDep(dependencyIds[i], AcDb::kForRead);

			HTREEITEM hAssocDepItem = addTreeItem(CConstrainDbgHelper::objToClassAndHandleStr(assocDep.object(), str), hActionBodyItem, dependencyIds[i]);
				
			AcDbCompoundObjectId compoundObjectId;
			if((es = assocDep->getDependentOnCompoundObject(compoundObjectId)) == Acad::eOk)
			{
				AcDbObjectIdArray fullPath;
				compoundObjectId.getFullPath(fullPath);

				for(int j = 0; j<fullPath.length(); ++j)
				{
					AcDbObjectPointer <AcDbObject> objDep(fullPath[j], AcDb::kForRead);

					HTREEITEM hObjDepItem = addTreeItem(CConstrainDbgHelper::objToClassAndHandleStr(objDep.object(), str), hAssocDepItem, fullPath[j]);
				}
			}
		}
	}

	return hSurfaceItem;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Use: Display a list of Name-Value pairs "ObjectProperty" in the ListView control
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSurfaceDbgDlg::displayPropertyList(vector<ObjectProperty>& properties)
{
	this->LockWindowUpdate();

    m_ListCtrl.DeleteAllItems();

    LV_ITEM lvi;
    lvi.mask = 0;
	lvi.iSubItem = 0;
    int row = 0;

	vector <ObjectProperty>::iterator iterator;

	for(iterator = properties.begin(); iterator != properties.end(); ++iterator)
	{
        lvi.iItem = row;
        
        m_ListCtrl.InsertItem(&lvi);
        m_ListCtrl.SetItemText(row, 0, iterator->name.c_str());
        m_ListCtrl.SetItemText(row, 1, iterator->value.c_str());
        ++row;
    }

	this->UnlockWindowUpdate();
}
