﻿// (C) Copyright 2002-20010 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// This sample demonstrate the use of the Associative Surfaces API
//
// Products: AutoCAD (and verticals) 2011 and higher
//
// Written by Philippe Leefsma - 2009/2010 - Developer Technical Services
// Philippe.Leefsma@Autodesk.com
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

using Autodesk.AutoCAD;
using Autodesk.AutoCAD.Runtime;
using Autodesk.AutoCAD.ApplicationServices;
using Autodesk.AutoCAD.DatabaseServices;
using Autodesk.AutoCAD.Geometry;
using Autodesk.AutoCAD.EditorInput;

[assembly: CommandClass(typeof(HammerSurfaces.AssocSurfaces))]

namespace HammerSurfaces
{
    public class AssocSurfaces : IExtensionApplication
    {
        bool AssociativeFlag = true;

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Associative Surfaces can only be created if system variable SURFACEASSOCIATIVITY is set to 1
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public void Initialize()
        {
            //Make sure Surface Associativity is enabled
            Application.SetSystemVariable("SURFACEASSOCIATIVITY", 1);
        }

        public void Terminate()
        {
            // Do plug-in application clean up here
        }

        [CommandMethod("CreateAll")]
        public void CreateAll()
        {
            CreateExtrudedSurface();
            CreateRevolvedSurface();
            CreateBlendSurface();
            CreateSweptSurface();
            CreateOffsetSurface();
            CreateNetworkSurface();
            CreateLoftSurface();
            CreatePatchSurface();
            CreateTrimmedSurface();

            Application.DocumentManager.MdiActiveDocument.SendStringToExecute("vpoint 0,5,-5 ", false, false, false);
            Application.DocumentManager.MdiActiveDocument.SendStringToExecute("zoom e ", false, false, false);
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Toggle AssociativeFlag: Controls if any of the command below will create an associative surface or not
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("ToggleAssociativeFlag")]
        public void ToggleAssociativeFlag()
        {
            AssociativeFlag = !AssociativeFlag;

            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            ed.WriteMessage("\nAssociative Flag is now turned " + (AssociativeFlag ? "ON" : "OFF"));
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Create ExtrudedSurface
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("CreateExtrudedSurface")]
        public void CreateExtrudedSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                //Create profile entity
                Circle circle = new Circle(new Point3d(15, 0, 15), new Vector3d(0, 1, 0), 3);

                //Profile entity needs to be DB resident
                AcHelper.PostToDatabase(circle);

                Profile3d profile = new Profile3d(circle);

                Vector3d direction = new Vector3d(0, 5, 0);

                //Generates SweepOptions object
                SweepOptions sweepOptions = new SweepOptions();

                //Create ExtrudedSurface 
                ObjectId surfId = Autodesk.AutoCAD.DatabaseServices.Surface.CreateExtrudedSurface(profile,
                    direction,
                    sweepOptions,
                    AssociativeFlag);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Create a RevolvedSurface
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("CreateRevolvedSurface")]
        public void CreateRevolvedSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                //Create profile entity
                Circle circle = new Circle(new Point3d(-25, 0, 20), new Vector3d(0, 1, 0), 2);

                //Profile entity needs to be DB resident
                AcHelper.PostToDatabase(circle);

                Profile3d profile = new Profile3d(circle);

                Point3d axisPoint = new Point3d(-20, 0, 20);
                Vector3d axis = new Vector3d(0, 0, 1);

                RevolveOptions revolveOptions = new RevolveOptions();

                //Create the RevolvedSurface
                ObjectId surfId = Autodesk.AutoCAD.DatabaseServices.Surface.CreateRevolvedSurface(profile,
                    axisPoint,
                    axis,
                    300 * Math.PI / 180,
                    0,
                    revolveOptions,
                    AssociativeFlag);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Create a BlendSurface
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("CreateBlendSurface")]
        public void CreateBlendSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                //Create profile entities
                Circle circle1 = new Circle(new Point3d(20, 0, -15), new Vector3d(0, 1, 0), 2);
                Circle circle2 = new Circle(new Point3d(25, 10, -15), new Vector3d(1, 0, 0), 5);

                //Profile entities need to be DB resident
                AcHelper.PostToDatabase(circle1);
                AcHelper.PostToDatabase(circle2);

                LoftProfile startProfile = new LoftProfile(circle1);
                LoftProfile endProfile = new LoftProfile(circle2);

                BlendOptions blendOptions = new BlendOptions();

                //Create the BlendSurface
                ObjectId surfId = Autodesk.AutoCAD.DatabaseServices.Surface.CreateBlendSurface(startProfile,
                    endProfile,
                    blendOptions,
                    AssociativeFlag);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Create a SweptSurface
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("CreateSweptSurface")]
        public void CreateSweptSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                //Create swept entity, a circle
                Circle sweepEntity = new Circle(new Point3d(15, 15, 0), new Vector3d(1, 0, 0), 1);

                //Create path entity, a pline
                Polyline pline = new Polyline();

                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(15, 15), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(10, 15), Math.Tan(Math.PI / 4), 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(10, 10), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(12, 10), -Math.Tan(Math.PI / 4), 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(12, 0), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(0, 0), 0, 0, 0);

                Matrix3d Transfo = Matrix3d.Displacement(new Vector3d(-20, 0, -20));

                sweepEntity.TransformBy(Transfo);
                pline.TransformBy(Transfo);

                //Profile entities need to be DB resident
                AcHelper.PostToDatabase(sweepEntity);
                AcHelper.PostToDatabase(pline);

                Profile3d sweepProfile = new Profile3d(sweepEntity);

                Profile3d pathProfile = new Profile3d(pline);

                //SweepOptions properties are read only, need to use SweepOptionsBuilder
                SweepOptionsBuilder sweepBuilder = new SweepOptionsBuilder();

                sweepBuilder.ScaleFactor = 4.0;

                //Generates SweepOptions object
                SweepOptions sweepOptions = sweepBuilder.ToSweepOptions();

                //Create SweptSurface
                ObjectId surfId = Autodesk.AutoCAD.DatabaseServices.Surface.CreateSweptSurface(sweepProfile,
                    pathProfile,
                    sweepOptions,
                    AssociativeFlag);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Create an OffsetSurface 
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("CreateOffsetSurface")]
        public void CreateOffsetSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                //First creates a DB resident Region based on a circle
                Circle circle = new Circle(new Point3d(20, 0, 0), new Vector3d(0, 1, 0), 5);
                
                DBObjectCollection curves = new DBObjectCollection();
                curves.Add(circle);

                DBObjectCollection regions = Region.CreateFromCurves(curves);

                Region region = regions[0] as Region;

                //Region needs to be DB resident
                AcHelper.PostToDatabase(region);

                //Explicitely dispose the non-db resident entities
                circle.Dispose();
                regions.Dispose();

                //Create the OffsetSurface
                ObjectId surfId = Autodesk.AutoCAD.DatabaseServices.Surface.CreateOffsetSurface(region, 
                    5.0, 
                    AssociativeFlag);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Create a NetworkSurface
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("CreateNetworkSurface")]
        public void CreateNetworkSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                //Create profile entities
                Arc arc1 = new Arc(new Point3d(-25, 0, 5), 5, 0, Math.PI);
                Arc arc2 = new Arc(new Point3d(-25, 0, -5), 5, 0, Math.PI);

                Line line1 = new Line(new Point3d(-20, 0, -5), new Point3d(-20, 0, 5));
                Line line2 = new Line(new Point3d(-30, 0, -5), new Point3d(-30, 0, 5));

                //All profile entities need to be db resident
                AcHelper.PostToDatabase(arc1);
                AcHelper.PostToDatabase(arc2);
                AcHelper.PostToDatabase(line1);
                AcHelper.PostToDatabase(line2);

                List<LoftProfile> uProfiles = new List<LoftProfile>();
                List<LoftProfile> vProfiles = new List<LoftProfile>();

                uProfiles.Add(new LoftProfile(arc1));
                uProfiles.Add(new LoftProfile(arc2));

                vProfiles.Add(new LoftProfile(line1));
                vProfiles.Add(new LoftProfile(line2));

                //Create the NetworkSurface
                ObjectId surfId = Autodesk.AutoCAD.DatabaseServices.Surface.CreateNetworkSurface(
                    uProfiles.ToArray(),
                    vProfiles.ToArray(),
                    AssociativeFlag);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        //
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("CreatePatchSurface")]
        public void CreatePatchSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                Polyline pline = new Polyline();

                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(0, 0), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(10, -2), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(20, 2), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(15, 10), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(20, 15), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(10, 20), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(0, 20), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(-10, 5), 0, 0, 0);

                pline.Closed = true;

                Matrix3d Transfo = Matrix3d.Displacement(new Vector3d(-15, 0, 30));

                pline.TransformBy(Transfo);

                AcHelper.PostToDatabase(pline);

                EdgeRef[] edges = new EdgeRef[] { new EdgeRef(pline) };

                PathRef path = new PathRef(edges);

                ObjectId surfId = Autodesk.AutoCAD.DatabaseServices.Surface.CreatePatchSurface(path, 
                    new ObjectIdCollection(), 
                    1, 
                    1, 
                    AssociativeFlag);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        //
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("CreateLoftSurface")]
        public void CreateLoftSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                Circle circle1 = new Circle(new Point3d(20, 0, 30), new Vector3d(0, 1, 0), 1);
                Circle circle2 = new Circle(new Point3d(30, 15, 30), new Vector3d(0, 1, 0), 7);

                Point3dCollection points = new Point3dCollection();
                points.Add(new Point3d(19, 0, 30));
                points.Add(new Point3d(26, 5, 30));
                points.Add(new Point3d(15, 10, 30));
                points.Add(new Point3d(23, 15, 30));

                Spline spline = new Spline(points, 5, 0.01);

                AcHelper.PostToDatabase(circle1);
                AcHelper.PostToDatabase(circle2);
                AcHelper.PostToDatabase(spline);

                List<LoftProfile> crossSections = new List<LoftProfile>();

                crossSections.Add(new LoftProfile(circle1));
                crossSections.Add(new LoftProfile(circle2));

                List<LoftProfile> guides = new List<LoftProfile>();

                guides.Add(new LoftProfile(spline));

                LoftOptions loftOptions = new LoftOptions();

                ObjectId surfId = Autodesk.AutoCAD.DatabaseServices.Surface.CreateLoftedSurface(
                    crossSections.ToArray(), 
                    guides.ToArray(), 
                    null, 
                    loftOptions, 
                    AssociativeFlag);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        //
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("CreateTrimmedSurface")]
        public void CreateTrimmedSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                Circle patchCircle = new Circle(new Point3d(45,0,0), new Vector3d(0,1,0), 6);

                AcHelper.PostToDatabase(patchCircle);

                EdgeRef[] edges = new EdgeRef[] { new EdgeRef(patchCircle) };

                PathRef path = new PathRef(edges);

                ObjectId patchId = Autodesk.AutoCAD.DatabaseServices.Surface.CreatePatchSurface(path, 
                    new ObjectIdCollection(), 
                    1, 
                    1, 
                    AssociativeFlag);

                //array of ids of cutting entities that their bodies 
                //will be directly used to trim the surface without further treatment such as projection.
                ObjectIdCollection toolSurfaceIds = new ObjectIdCollection();

                //array of ids of cutting curves that will be used 
                //to trim the surface by first projecting them to the surface  
                ObjectIdCollection toolCurveIds = new ObjectIdCollection();

                Arc arc = new Arc(new Point3d(45, 0, 6), new Vector3d(0, 1, 0), 5, Math.PI, 0);

                toolCurveIds.Add(AcHelper.PostToDatabase(arc));
               
                //array of projection direction of each cutting curve in toolCurveIds, 
                //so the length of projVectors should be equal to the length of toolCurveIds  
                Vector3dCollection projVectors = new Vector3dCollection();
                projVectors.Add(patchCircle.Normal);
               
                //pick point that is used to specify which area of a surface should be trimmed
                Point3d pickPoint = arc.Center;

                //when using pick point to find out which area to trim, 
                //caller should also specify the view direction which will be used 
                //to form a ray starting from the pick point in the view direction 
                //and see which area is hit by the ray first.  
                Vector3d viewVector = patchCircle.Normal;

                //TrimSurface returns void since surfaces are already db resident
                Autodesk.AutoCAD.DatabaseServices.Surface.TrimSurface(patchId,
                    toolSurfaceIds,
                    toolCurveIds,
                    projVectors,
                    pickPoint,
                    viewVector,
                    true,
                    AssociativeFlag);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Dump the selected surface's ActionBody if surface is associative
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("ReadSurfaceActionBody")]
        public void ReadSurfaceActionBody()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptEntityOptions peo = new PromptEntityOptions("\nSelect Associative Surface: ");
            peo.SetRejectMessage("\nInvalid Selection...");
            peo.AddAllowedClass(typeof(Autodesk.AutoCAD.DatabaseServices.Surface), false);

            PromptEntityResult per = ed.GetEntity(peo);

            if (per.Status != PromptStatus.OK) 
                return;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                Autodesk.AutoCAD.DatabaseServices.Surface surface = Tx.GetObject(per.ObjectId, OpenMode.ForRead) 
                    as Autodesk.AutoCAD.DatabaseServices.Surface;

                //Make sure the surface is associative
                if (surface.CreationActionBodyId == ObjectId.Null)
                {
                    ed.WriteMessage("\nCreationActionBodyId = ObjectId.Null");
                    return;
                }

                AssocActionBody actionBody = Tx.GetObject(surface.CreationActionBodyId, OpenMode.ForRead) 
                    as AssocActionBody;

                ed.WriteMessage("\n-- " + actionBody.ToString() + " --");

                //All surface ActionBodies are ParamBased, so we don't need to check here
                ReadParamBasedActionBody(actionBody as AssocParamBasedActionBody);

                //Check if ActionBody is PathBased
                if (actionBody is AssocPathBasedSurfaceActionBody)
                {
                    ReadPathBasedActionBody(actionBody as AssocPathBasedSurfaceActionBody);
                }
                
                //All ActionBodies have dependencies
                ReadDependencies(actionBody);
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Dump ActionBody Parameters: ALL surface actionbodies are ParamBased
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        void ReadParamBasedActionBody(AssocParamBasedActionBody paramBasedActionBody)
        {
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            ed.WriteMessage("\n.\n--------- Parameters ---------");

            foreach (ObjectId paramId in paramBasedActionBody.OwnedParams)
            {
                AssocActionParam param = paramId.GetObject(OpenMode.ForRead) as AssocActionParam;

                ed.WriteMessage("\n Param: " + param.ToString());
            }

            ed.WriteMessage("\n.\n--------- Value Parameters ---------");

            foreach (string paramName in paramBasedActionBody.OwnedValueParamNames())
            {
                AssocParamBasedActionBody.ValueParam vParam = paramBasedActionBody.GetValueParam(paramName);

                string value = vParam.Expression;

                if (value == "")
                { 
                    ResultBufferEnumerator rbEnum = vParam.Value.GetEnumerator();

                    if (rbEnum.MoveNext()) 
                        value = rbEnum.Current.Value.ToString();
                }

                ed.WriteMessage("\n " + paramName + " = " + value);
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Dump a PathBased Surface ActionBody: Not all surface actionbodies are PathBased
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        void ReadPathBasedActionBody(AssocPathBasedSurfaceActionBody pathBasedSurfaceActionBody)
        {
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            ed.WriteMessage("\n.\n--------- PathBased ActionBody ---------");

            foreach (ObjectId paramId in pathBasedSurfaceActionBody.InputPathParams)
            {
                ed.WriteMessage("\n InputPathParam:");

                AssocPathActionParam pathParam = paramId.GetObject(OpenMode.ForRead) as AssocPathActionParam;

                EdgeRef[][] edgeRefArrays = pathParam.GetEdgeRefArray();
                
                foreach (EdgeRef[] edgeRefs in edgeRefArrays)
                {
                    foreach (EdgeRef edgeRef in edgeRefs)
                    {
                        Curve3d curve = edgeRef.Curve;
                        
                        Entity entityRef = edgeRef.CreateEntity();

                        ed.WriteMessage("\n    Param Curve: " + curve.ToString());
                        ed.WriteMessage("\n    Param Entity Ref: " + entityRef.ToString());

                        entityRef.Dispose();
                    }
                }
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Dump ActionBody dependencies: ALL surface actionbodies have Dependencies
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        void ReadDependencies(AssocActionBody actionBody)
        {
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            ed.WriteMessage("\n.\n--------- Dependencies ---------");

            foreach (ObjectId id in actionBody.GetDependencies(true, true))
            {
                ed.WriteMessage("\n Dependency:");

                AssocDependency assocDep = id.GetObject(OpenMode.ForRead) as AssocDependency;

                string nature = "";

                if (assocDep.IsReadDependency && assocDep.IsWriteDependency)
                {
                    nature = "Read/Write";
                }
                else if (assocDep.IsReadDependency)
                {
                    nature = "Read";
                }
                else if (assocDep.IsWriteDependency)
                {
                    nature = "Read/Write";
                }

                ed.WriteMessage("\n     Nature: " + nature);

                CompoundObjectId compoundObjectId = assocDep.DependentOnCompoundObject;

                foreach (ObjectId pathId in compoundObjectId.FullPath)
                {
                    DBObject obj = pathId.GetObject(OpenMode.ForRead);

                    ed.WriteMessage("\n     Object: " + obj.ToString());
                    ed.WriteMessage("\n     ObjectId: " + obj.ObjectId.ToString()); 
                }
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Modify Height and Base radius of the ExtrudedSurface created above
        //
        // Nb: Base radius will work only if the surface is extruded from a circle
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("ModifyExtrudedSurface")]
        public void ModifyExtrudedSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptEntityOptions peo = new PromptEntityOptions("\nSelect ExtrudedSurface: ");
            peo.SetRejectMessage("\nInvalid Selection...");
            peo.AddAllowedClass(typeof(ExtrudedSurface), true);

            PromptEntityResult per = ed.GetEntity(peo);

            if (per.Status != PromptStatus.OK) 
                return;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                Autodesk.AutoCAD.DatabaseServices.Surface surface = Tx.GetObject(per.ObjectId, OpenMode.ForRead) 
                    as Autodesk.AutoCAD.DatabaseServices.Surface;

                //Check that the surface is associative
                //If yes "surface.CreationActionBodyId" will not be null
                if (surface.CreationActionBodyId == ObjectId.Null)
                {
                    ed.WriteMessage("\nNot an Associative Surface...");
                    return;
                }
                 
                //Open ActionBody
                AssocExtrudedSurfaceActionBody actionBody = Tx.GetObject(surface.CreationActionBodyId, 
                    OpenMode.ForWrite) as AssocExtrudedSurfaceActionBody;

                PromptResult psr1 = ed.GetString("\nEnter Extrusion Height Expression: ");

                if (psr1.Status != PromptStatus.OK)
                    return;

                actionBody.SetHeight(0, psr1.StringResult, "");

                PromptResult psr2 = ed.GetString("\nEnter Extrusion TaperAngle Expression (degrees): ");

                if (psr2.Status != PromptStatus.OK)
                    return;

                actionBody.SetTaperAngle(0, psr2.StringResult, "");


                //Search in ActionBody dependencies to find the base circle
                //If found, prompt user for a new radius and update circle
                foreach (ObjectId id in actionBody.GetDependencies(true, true))
                {
                    AssocDependency assocDep = id.GetObject(OpenMode.ForRead) as AssocDependency;

                    CompoundObjectId compoundObjectId = assocDep.DependentOnCompoundObject;

                    foreach (ObjectId pathId in compoundObjectId.FullPath)
                    {
                        DBObject obj = pathId.GetObject(OpenMode.ForRead);

                        if (obj is Circle)
                        {
                            Circle baseCircle = obj as Circle;

                            baseCircle.UpgradeOpen();

                            PromptDoubleOptions pdo = new PromptDoubleOptions("\nEnter base Radius: ");
                            pdo.AllowNegative = false;

                            PromptDoubleResult pdr = ed.GetDouble(pdo);

                            if (pdr.Status != PromptStatus.OK) return;

                            baseCircle.Radius = pdr.Value;
                        }
                    }
                }

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Select an OffsetSurface 
        // then add an associative variable "OFFSET1" and affect it to the offset parameter.
        //
        // Once the offset surface is created, use command "AddModifAssocVar" to modify "OFFSET1" value.
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("AddAssocVarToSurface")]
        public void AddAssocVarToSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptEntityOptions peo = new PromptEntityOptions("\nSelect Offset Surface: ");
            peo.SetRejectMessage("\nInvalid Selection...");
            peo.AddAllowedClass(typeof(Autodesk.AutoCAD.DatabaseServices.Surface), false);

            PromptEntityResult per = ed.GetEntity(peo);

            if (per.Status != PromptStatus.OK) 
                return;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                Autodesk.AutoCAD.DatabaseServices.Surface surface = Tx.GetObject(per.ObjectId, OpenMode.ForRead) 
                    as Autodesk.AutoCAD.DatabaseServices.Surface;

                //Check that the surface is associative
                //If yes "surface.CreationActionBodyId" will not be null
                if (surface.CreationActionBodyId == ObjectId.Null)
                {
                    ed.WriteMessage("\nNot an Associative Surface...");
                    return;
                }

                //Open offset ActionBody
                AssocOffsetSurfaceActionBody actionBody = Tx.GetObject(surface.CreationActionBodyId, 
                    OpenMode.ForRead) as AssocOffsetSurfaceActionBody;

                if(actionBody == null)
                {
                    ed.WriteMessage("\nNot an Offset Surface...");
                    return;
                }

                actionBody.UpgradeOpen();

                //Add new "OFFSET1" using our helper method
                ObjectId varId = AcHelper.AddOrModifVariable("OFFSET1", "10");

                actionBody.SetDistance(0, "OFFSET1", "");

                //Old Style
                //Look for "offsetdistance" parameter and affect its "Expression" with "OFFSET1"
                /*foreach (string paramName in actionBody.OwnedValueParamNames())
                {
                    if (paramName.ToLower() == "offsetdistance")
                    {
                        AssocParamBasedActionBody.ValueParam vParam = actionBody.GetValueParam(paramName);

                        vParam.Expression = "OFFSET1";

                        actionBody.SetValueParam(paramName, vParam, true);
                        break;
                    }
                }*/

                Tx.Commit();
            }               
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Create 4 connected associative surfaces
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("CreateVase")]
        public void CreateVase()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                //Create 1st Surface: a RevolvedSurface based on polyline

                Polyline pline = new Polyline();

                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(5, 0), Math.Tan((100 * Math.PI/180) / 4), 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(5, 10), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(4, 10), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(4, 11), Math.Tan((-60 * Math.PI / 180) / 4), 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(4, 15), 0, 0, 0);
                pline.AddVertexAt(pline.NumberOfVertices, new Point2d(1.5, 16), 0, 0, 0);

                AcHelper.PostToDatabase(pline);

                Point3d axisPt = new Point3d(0, 0, 0);
                Vector3d axis = new Vector3d(0, 1, 0);

                RevolveOptions revolveOptions = new RevolveOptions();

                Profile3d profileRev = new Profile3d(pline);

                ObjectId revId = Autodesk.AutoCAD.DatabaseServices.Surface.CreateRevolvedSurface(profileRev,
                    axisPt,
                    axis,
                    2 * Math.PI,
                    0,
                    revolveOptions,
                    AssociativeFlag);


                //Retrieve edge from revolved surface and create 2nd Surface: an ExtrudedSurface

                Entity revSurface = Tx.GetObject(revId, OpenMode.ForWrite) as Entity;

                revSurface.ColorIndex = 1;

                FullSubentityPath[] edges1 = AcHelper.GetEdgeSubentities(revSurface);

                FullSubentityPath subentPath1 = FullSubentityPath.Null;
                FullSubentityPath subentPath2 = FullSubentityPath.Null;

                //Loop through the RevolvedSurface edges and try to locate the top edge circle
                foreach (FullSubentityPath subentPath in edges1)
                {
                    Entity subEntity = revSurface.GetSubentity(subentPath);

                    if (subEntity is Circle)
                    {
                        Circle subCircle = subEntity as Circle;

                        if (Math.Abs(subCircle.Center.Y) < 0.001)
                        {
                            subentPath1 = subentPath;
                        }

                        if (Math.Abs(subCircle.Center.Y - 16.0) < 0.001)
                        {
                            subentPath2 = subentPath;
                        }
                    }
                }

                PathRef path = AcHelper.CreatePathRef(subentPath1);

                ObjectId patchId = Autodesk.AutoCAD.DatabaseServices.Surface.CreatePatchSurface(path,
                    new ObjectIdCollection(),
                    0,
                    0,
                    AssociativeFlag);


                Entity patchSurface = Tx.GetObject(patchId, OpenMode.ForWrite) as Entity;

                patchSurface.ColorIndex = 2;

                //Create a profile using the FullSubentityPath of the circle edge
                Profile3d profileExt = AcHelper.CreateProfile3d(subentPath2);

                Vector3d dirExt = new Vector3d(0, 4, 0);

                SweepOptions sweepOptions1 = new SweepOptions();

                //Create 2nd surface based on the edge profile
                ObjectId extrId = Autodesk.AutoCAD.DatabaseServices.Surface.CreateExtrudedSurface(profileExt,
                    dirExt,
                    sweepOptions1,
                    AssociativeFlag);

                Entity extrSurface = Tx.GetObject(extrId, OpenMode.ForWrite) as Entity;
                
                extrSurface.ColorIndex = 3;



                FullSubentityPath[] edges2 = AcHelper.GetEdgeSubentities(extrSurface);

                FullSubentityPath subentPath3 = FullSubentityPath.Null;

                //Loop through the ExtrudedSurface edges and try to locate the top edge circle
                foreach (FullSubentityPath subentPath in edges2)
                {
                    Entity subEntity = extrSurface.GetSubentity(subentPath);

                    if (subEntity is Circle)
                    {
                        Circle subCircle = subEntity as Circle;

                        if (Math.Abs(subCircle.Center.Y - 20.0) < 0.001)
                        {
                            subentPath3 = subentPath;
                            break;
                        }
                    }
                }

                Circle circle = new Circle(new Point3d(0, 25, 0), new Vector3d(0, 1, 0), 3);
                AcHelper.PostToDatabase(circle);
               
                List<LoftProfile> crossSections = new List<LoftProfile>();

                crossSections.Add(new LoftProfile(circle));                
                crossSections.Add(new LoftProfile(AcHelper.CreatePathRef(subentPath3)));

                ObjectId blendId = Autodesk.AutoCAD.DatabaseServices.Surface.CreateBlendSurface(
                    crossSections[0], 
                    crossSections[1],
                    new BlendOptions(),
                    AssociativeFlag);

                Entity blendSurface = Tx.GetObject(blendId, OpenMode.ForWrite) as Entity;

                blendSurface.ColorIndex = 4;
                
               
                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Project Entity to a Surface
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("ProjectOnToSurface")]
        public void ProjectOnToSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptEntityOptions peo1 = new PromptEntityOptions("\nSelect Surface: ");
            peo1.SetRejectMessage("\nInvalid Selection...");
            peo1.AddAllowedClass(typeof(Autodesk.AutoCAD.DatabaseServices.Surface), false);

            PromptEntityResult per = ed.GetEntity(peo1);

            if (per.Status != PromptStatus.OK) 
                return;

            PromptEntityOptions peo2 = new PromptEntityOptions("\nSelect Entity: ");
            peo2.SetRejectMessage("\nInvalid Selection...");
            peo2.AddAllowedClass(typeof(Entity), false);

            PromptEntityResult per2 = ed.GetEntity(peo2);

            if (per2.Status != PromptStatus.OK) 
                return;

            PromptPointOptions ppo = new PromptPointOptions("\nSpecify direction: ");
            ppo.BasePoint = Point3d.Origin;
            ppo.UseBasePoint = true;
            ppo.UseDashedLine = true;
            
            PromptPointResult ppr = ed.GetPoint(ppo);

            if (ppr.Status != PromptStatus.OK)
                return;

            Vector3d direction = ppr.Value.GetAsVector();

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                Autodesk.AutoCAD.DatabaseServices.Surface surface = Tx.GetObject(per.ObjectId, OpenMode.ForRead) 
                    as Autodesk.AutoCAD.DatabaseServices.Surface;

                Entity entity = Tx.GetObject(per2.ObjectId, OpenMode.ForRead) as Entity;

                Entity[] results = surface.ProjectOnToSurface(entity, direction);

                foreach (Entity result in results)
                {
                    result.ColorIndex = 1;

                    AcHelper.PostToDatabase(result);
                }

                Tx.Commit();
            }
        }
    }
}