// (C) Copyright 2002-20010 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// This sample demonstrate the use of the NURBS Surfaces API
//
// Products: AutoCAD (and verticals) 2011 and higher
//
// Written by Philippe Leefsma - 2009/2010 - Developer Technical Services
// Philippe.Leefsma@Autodesk.com
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using Autodesk.AutoCAD.Runtime;
using Autodesk.AutoCAD.Geometry;
using Autodesk.AutoCAD.EditorInput;
using Autodesk.AutoCAD.DatabaseServices;
using Autodesk.AutoCAD.ApplicationServices;

[assembly: CommandClass(typeof(HammerSurfaces.NurbSurfaces))]

namespace HammerSurfaces
{
    public class NurbSurfaces
    {
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Use: Generates a Nurbs surface from a Surface using "ConvertToNurbSurface" method
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("NurbFromSurface")]
        public void NurbFromSurface()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptEntityOptions peo = new PromptEntityOptions("\nSelect Surface: ");
            peo.SetRejectMessage("\nInvalid Selection...");
            peo.AddAllowedClass(typeof(Autodesk.AutoCAD.DatabaseServices.Surface), false);

            PromptEntityResult per = ed.GetEntity(peo);

            if (per.Status != PromptStatus.OK)
                return;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                Autodesk.AutoCAD.DatabaseServices.Surface surface = Tx.GetObject(per.ObjectId, OpenMode.ForWrite) 
                    as Autodesk.AutoCAD.DatabaseServices.Surface;

                Autodesk.AutoCAD.DatabaseServices.NurbSurface[] nurbs = surface.ConvertToNurbSurface();

                foreach (Autodesk.AutoCAD.DatabaseServices.NurbSurface nurb in nurbs)
                {
                    AcHelper.PostToDatabase(nurb);
                }

                surface.Erase(true);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Create a NURB Surface
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public ObjectId CreateNurbSurface(Database db, string nurbData)
        {
            nurbData.Replace(" ", "");

            string[] inputData = nurbData.Split(',');

            int idx = 0;

            int uDegree = int.Parse(inputData[idx++]);
            int vDegree = int.Parse(inputData[idx++]);

            bool rational = bool.Parse(inputData[idx++]);

            int uNumControlPoints = int.Parse(inputData[idx++]);
            int vNumControlPoints = int.Parse(inputData[idx++]);

            Point3dCollection ctrlPts = new Point3dCollection();

            for (int i = 0; i < uNumControlPoints * vNumControlPoints; ++i)
            {
                ctrlPts.Add(new Point3d(double.Parse(inputData[idx++]),
                                        double.Parse(inputData[idx++]),
                                        double.Parse(inputData[idx++])));
            }

            DoubleCollection weights = new DoubleCollection();

            for (int i = 0; i < uNumControlPoints * vNumControlPoints; ++i)
            {
                weights.Add(double.Parse(inputData[idx++]));
            }
            
            KnotCollection uknots = new KnotCollection();

            for (int i = 0; i < (uNumControlPoints + uDegree + 1); ++i)
            {
                uknots.Add(double.Parse(inputData[idx++]));
            }

            KnotCollection vknots = new KnotCollection();

            for (int i = 0; i < (vNumControlPoints + vDegree + 1); ++i)
            {
                vknots.Add(double.Parse(inputData[idx++]));
            }

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {                 
                Autodesk.AutoCAD.DatabaseServices.NurbSurface nurb = 
                    new Autodesk.AutoCAD.DatabaseServices.NurbSurface(uDegree, 
                        vDegree, 
                        rational, 
                        uNumControlPoints, 
                        vNumControlPoints, 
                        ctrlPts, 
                        weights, 
                        uknots, 
                        vknots);

                nurb.SetDatabaseDefaults();

                ObjectId nurbId = AcHelper.PostToDatabase(nurb);

                Tx.Commit();

                return nurbId;
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        //  This command will add nurbsurface based on the passed-in parameters     
        //  for Autodesk.AutoCAD.DatabaseServices.NurbSurface
        //
        //  1. uDegree
        //  2. VDegree
        //  3. rational
        //  4. uNumControlPoints
        //  5. vNumControlPoints
        //  6. Control Points
        //  7. Weights
        //  8. Uknots
        //  9. VKnots      
        //
        // The control points are contained in an array of AcGePoint3d. 
        // The v index varies first. 
        // That is, a column of v control points for the first u value is followed by the row of v control points 
        // for the next u value, e.g., [0,0], [0,1], ...[0,n], ...,  [m,0], [m,1], ...[m,n]
        // The weights array is required when rational == true. 
        // And it must have the same number of entries as the ctrlPts array. 
        // All weight values must be greater than 0.0. 
        // The values in this array are in the same sequential order as the control points. 
        // When rational == false, the passed-in weight array should be empty.
        // uKnots array must have (uNumControlPoints + uDegree + 1) entries 
        // vKnots array must have (vNumControlPoints + vDegree + 1) entries. 
        // Each knot must have a value greater than or equal to the value of its predecessor in the array
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("NurbCreate")]
        public void NurbCreate()
        {
            Database db = HostApplicationServices.WorkingDatabase;

            string nurb1 = "1,1,False,2,2,0,0,0,10,0,0,0,10,0,10,10,0,0,0,0,0,0,0,10,10,0,0,10,10";

            string nurb2 = "3,1,False,7,2,9.04,0.20,0.00,4.64,0.98,3.01,7.48,4.00,0.00,6.66,4.97,3.01,8.60,7.92,0.00,8.12,8.82,3.01,16.76,4.56,0.00,15.14,6.44,3.01,16.16,11.39,0.00,16.16,11.39,3.01,18.53,12.48,-2.27,18.53,12.48,4.07,21.06,11.20,0.00,20.07,13.98,3.01,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,6.25,11.43,16.29,19.02,19.02,19.02,19.02,0.00,0.00,3.01,3.01";

            string nurb3 = "2,2,True,9,3,21.46,8.16,0.00,20.85,7.14,-1.97,18.08,7.14,-3.21,19.88,6.78,0.00,19.88,6.78,-1.51,18.71,6.78,-2.45,20.06,5.76,0.00,20.06,5.76,-1.59,18.82,5.76,-2.59,20.25,4.74,0.00,20.25,4.74,-1.68,18.71,4.23,-2.74,21.28,3.84,0.00,21.28,4.74,-2.18,18.50,4.33,-3.54,22.32,4.74,0.00,22.32,4.74,-2.67,19.46,4.74,-4.35,22.50,5.76,0.00,22.50,5.76,-2.76,20.35,5.76,-4.49,22.68,6.78,0.00,22.68,6.78,-2.85,20.46,6.78,-5.10,21.71,7.14,0.00,21.71,7.14,-2.38,20.33,7.97,-3.88,1.00,0.90,1.00,0.77,0.69,0.77,1.00,0.90,1.00,0.77,0.69,0.77,1.00,0.90,1.00,0.77,0.69,0.77,1.00,0.90,1.00,0.77,0.69,0.77,1.00,0.90,1.00,-4.36,-4.36,-4.36,-2.96,-2.96,-1.57,-1.57,-0.18,-0.18,1.22,1.22,1.22,0.00,0.00,0.00,0.89,0.89,0.89";
                                           
            string nurb4 = "3,2,True,6,3,36.26,13.60,0.00,36.26,13.60,-1.53,35.19,13.60,-2.60,37.21,12.59,0.00,37.21,12.59,-1.92,35.85,12.59,-3.27,39.27,10.39,0.00,39.27,10.39,-2.77,37.31,10.39,-4.73,35.91,5.51,0.00,35.91,5.51,-1.38,34.93,5.51,-2.35,37.39,1.92,0.00,37.39,1.92,-1.99,35.98,1.92,-3.40,38.27,-0.23,0.00,38.27,-0.23,-2.36,36.61,-0.23,-4.02,1.00,0.92,1.00,1.00,0.92,1.00,1.00,0.92,1.00,1.00,0.92,1.00,1.00,0.92,1.00,1.00,0.92,1.00,0.00,0.00,0.00,0.00,3.71,8.13,14.71,14.71,14.71,14.71,0.00,0.00,0.00,0.79,0.79,0.79";
                                               
            CreateNurbSurface(db, nurb1);
            CreateNurbSurface(db, nurb2);
            CreateNurbSurface(db, nurb3);
            CreateNurbSurface(db, nurb4);

            Application.DocumentManager.MdiActiveDocument.SendStringToExecute("vpoint 1,1,1 ", false, false, false);
            Application.DocumentManager.MdiActiveDocument.SendStringToExecute("zoom e ", false, false, false);
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Use: Illustrates analyzing some properties of an existing Nurbs Surface
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("NurbAnalyze")]
        public void NurbAnalyze()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptEntityOptions peo = new PromptEntityOptions("\nSelect NURB Surface: ");
            peo.SetRejectMessage("\nInvalid Selection...");
            peo.AddAllowedClass(typeof(Autodesk.AutoCAD.DatabaseServices.NurbSurface), false);

            PromptEntityResult per = ed.GetEntity(peo);

            if (per.Status != PromptStatus.OK)
                return;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                Autodesk.AutoCAD.DatabaseServices.NurbSurface nurb = Tx.GetObject(per.ObjectId, OpenMode.ForRead)
                    as Autodesk.AutoCAD.DatabaseServices.NurbSurface;

                Point3d ptOnSurface;
                Vector3d normal;

                bool isPlanar = nurb.IsPlanar(out ptOnSurface, out normal);

                ed.WriteMessage("\nIs Planar: " + isPlanar.ToString());

                ed.WriteMessage("\nPicked point: " + per.PickedPoint.ToString());

                Point3d analyzedPoint = per.PickedPoint;

                if (!nurb.IsPointOnSurface(analyzedPoint))
                {
                    ed.WriteMessage("\nPicked point not on surface.");
                    ed.WriteMessage("\nUsing other point on surface: " + ptOnSurface.ToString());
                    analyzedPoint = ptOnSurface;
                }

                double u, v;
                nurb.GetParameterOfPoint(analyzedPoint, out u, out v);

                ed.WriteMessage("\nParameters [{0}, {1}]", u, v);

                normal = nurb.GetNormal(u, v);

                ed.WriteMessage("\nnormal at [{0}, {1}]: " + normal.ToString(), u, v);

                Tx.Commit();
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Use: Visualizes ISO Lines of the selected Nurb Surface
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        void IsoLines(ObjectId nurbId, int nbLinesInU, int nbLinesInV)
        {
            using (Transaction Tx = nurbId.Database.TransactionManager.StartTransaction())
            {
                Autodesk.AutoCAD.DatabaseServices.NurbSurface nurb = Tx.GetObject(nurbId, OpenMode.ForWrite)
                    as Autodesk.AutoCAD.DatabaseServices.NurbSurface;

                double uStep = Math.Abs(nurb.UKnots.EndParameter - nurb.UKnots.StartParameter) /
                    (double)(nbLinesInU - 1);

                for (double u = nurb.UKnots.StartParameter; u <= nurb.UKnots.EndParameter; u += uStep)
                {
                    Curve[] curves = nurb.GetIsolineAtU(u);

                    for (int i = 0; i < curves.Length; i++)
                    {
                        curves[i].ColorIndex = 1;

                        AcHelper.PostToDatabase(curves[i]);
                    }
                }

                double vStep = Math.Abs(nurb.VKnots.EndParameter - nurb.VKnots.StartParameter) /
                    (double)(nbLinesInV - 1);

                for (double v = nurb.VKnots.StartParameter; v <= nurb.VKnots.EndParameter; v += vStep)
                {
                    Curve[] curves = nurb.GetIsolineAtV(v);

                    for (int i = 0; i < curves.Length; i++)
                    {
                        curves[i].ColorIndex = 2;

                        AcHelper.PostToDatabase(curves[i]);
                    }
                }

                Tx.Commit();
            }
        }

        [CommandMethod("NurbIsoLines")]
        public void NurbIsoLines()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptEntityOptions peo = new PromptEntityOptions("\nSelect NURB Surface: ");
            peo.SetRejectMessage("\nInvalid Selection...");
            peo.AddAllowedClass(typeof(Autodesk.AutoCAD.DatabaseServices.NurbSurface), false);

            PromptEntityResult per = ed.GetEntity(peo);

            if (per.Status != PromptStatus.OK)
                return;

            PromptIntegerOptions pio1 = new PromptIntegerOptions("\nEnter number of IsoLines in U: ");
            pio1.AllowNegative = false;
            pio1.AllowZero = false;
            pio1.DefaultValue = 10;
            pio1.UseDefaultValue = true;

            PromptIntegerResult pir1 = ed.GetInteger(pio1);

            if (pir1.Status != PromptStatus.OK)
                return;

            PromptIntegerOptions pio2 = new PromptIntegerOptions("\nEnter number of IsoLines in V: ");
            pio2.AllowNegative = false;
            pio2.AllowZero = false;
            pio2.DefaultValue = 10;
            pio2.UseDefaultValue = true;

            PromptIntegerResult pir2 = ed.GetInteger(pio2);

            if (pir2.Status != PromptStatus.OK)
                return;

            IsoLines(per.ObjectId, pir1.Value, pir2.Value);
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Use: Offsets control Points of the selected Nurb
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        void OffsetCtrlPnts(ObjectId nurbId, Vector3d offset)
        {
            using (Transaction Tx = nurbId.Database.TransactionManager.StartTransaction())
            {
                Autodesk.AutoCAD.DatabaseServices.NurbSurface nurb = Tx.GetObject(nurbId, OpenMode.ForWrite)
                    as Autodesk.AutoCAD.DatabaseServices.NurbSurface;

                for (int u = 0; u < nurb.NumberOfControlPointsInU / 2; ++u)
                {
                    for (int v = 0; v < nurb.NumberOfControlPointsInV / 2; ++v)
                    {
                        Point3d ctrlPnt = nurb.GetControlPointAt(u, v);

                        try
                        {
                            nurb.SetControlPointAt(u, v, ctrlPnt.Add(offset));
                        }
                        catch
                        {
                            continue;
                        }
                    }
                }

                Tx.Commit();
            }
        }

        [CommandMethod("NurbOffsetCtrlPnts")]
        public void NurbOffsetCtrlPnts()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptEntityOptions peo = new PromptEntityOptions("\nSelect NURB Surface: ");
            peo.SetRejectMessage("\nInvalid Selection...");
            peo.AddAllowedClass(typeof(Autodesk.AutoCAD.DatabaseServices.NurbSurface), false);

            PromptEntityResult per = ed.GetEntity(peo);

            if (per.Status != PromptStatus.OK)
                return;

            OffsetCtrlPnts(per.ObjectId, new Vector3d(5, 0, 0));
        }       

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Use: Illustrates use of the "NurbSurface.Rebuild" Method
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        void NurbRebuild(ObjectId nurbId)
        {
            using (Transaction Tx = nurbId.Database.TransactionManager.StartTransaction())
            {
                Autodesk.AutoCAD.DatabaseServices.NurbSurface nurb = Tx.GetObject(nurbId, OpenMode.ForWrite) 
                    as Autodesk.AutoCAD.DatabaseServices.NurbSurface;

                nurb.Rebuild(nurb.DegreeInU + 1, 
                    nurb.DegreeInV + 1, 
                    2 * nurb.NumberOfControlPointsInU, 
                    2 * nurb.NumberOfControlPointsInV);

                Tx.Commit();
            }
        }

        [CommandMethod("NurbRebuild")]
        public void NurbRebuild()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptEntityOptions peo = new PromptEntityOptions("\nSelect NURB Surface: ");
            peo.SetRejectMessage("\nInvalid Selection...");
            peo.AddAllowedClass(typeof(Autodesk.AutoCAD.DatabaseServices.NurbSurface), false);

            PromptEntityResult per = ed.GetEntity(peo);

            if (per.Status != PromptStatus.OK) 
                return;

            NurbRebuild(per.ObjectId);
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Use: Illustrates use of the "NurbSurface.SetWeight" method
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        void ScaleWeights(ObjectId nurbId, double factor)
        {
            using (Transaction Tx = nurbId.Database.TransactionManager.StartTransaction())
            {
                Autodesk.AutoCAD.DatabaseServices.NurbSurface nurb = Tx.GetObject(nurbId, OpenMode.ForWrite) 
                    as Autodesk.AutoCAD.DatabaseServices.NurbSurface;

                for (int u = 0; u < nurb.NumberOfControlPointsInU; ++u)
                {
                    for (int v = 0; v < nurb.NumberOfControlPointsInV; ++v)
                    {
                        nurb.SetWeight(u, v, nurb.GetWeight(u, v) * factor);
                    }
                }

                Tx.Commit();
            }
        }

        [CommandMethod("NurbScaleWeights")]
        public void NurbScaleWeights()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptEntityOptions peo = new PromptEntityOptions("\nSelect NURB Surface: ");
            peo.SetRejectMessage("\nInvalid Selection...");
            peo.AddAllowedClass(typeof(Autodesk.AutoCAD.DatabaseServices.NurbSurface), false);

            PromptEntityResult per = ed.GetEntity(peo);

            if (per.Status != PromptStatus.OK) 
                return;

            ScaleWeights(per.ObjectId, 2.0);
        }

        
    }
}